import unittest
import tempfile
import os

from pixelgreat import helpers


class TestHelpers(unittest.TestCase):
    def test_strip_all(self):
        # 1) Just run a few samples to verify it works (simple behavior)
        vals = (
            (
                "   \n  Hello, world!  \n ",
                "Hello, world!"
            ),
            (
                "      \r\n  Hello, world!  \r\n ",
                "Hello, world!"
            ),
            (
                "   \r\n  Hello, world!  \n ",
                "Hello, world!"
            ),
            (
                "    Hello, world!  \r\n ",
                "Hello, world!"
            ),
            (
                "Hello, world!  \r\n ",
                "Hello, world!"
            ),
            (
                "  Hello, world!  ",
                "Hello, world!"
            )
        )

        for test_vals in vals:
            self.assertEqual(helpers.strip_all(test_vals[0]), test_vals[1])

    def test_file_path(self):
        # 1) Make some temporary files and verify we can see them
        temp_files = list()
        for x in range(30):
            temp_files.append(tempfile.NamedTemporaryFile())

        for file in temp_files:
            self.assertEqual(helpers.file_path(file.name), file.name)

        # 2) Delete the temporary files and verify we get a FileNotFound error
        for file in temp_files:
            file.close()

        for file in temp_files:
            self.assertRaises(FileNotFoundError, helpers.file_path, file.name)

    def test_clip(self):
        # 1) Just test a bunch of values to verify it works
        vals = (
            (
                (1, 0, 1),
                1
            ),
            (
                (1, 0, 0.5),
                .5
            ),
            (
                (0, 0, 0.5),
                0
            ),
            (
                (-4, 0, 0.5),
                0
            ),
            (
                (100, 0, 12.345),
                12.345
            )
        )

        for test_vals in vals:
            self.assertEqual(
                helpers.clip(test_vals[0][0], test_vals[0][1], test_vals[0][2]),
                test_vals[1]
            )

    def test_get_centered_dimensions(self):
        pass

    def test_round_to_division(self):
        pass

    def test_tile_image(self):
        pass

    def test_lighten_image(self):
        pass


if __name__ == '__main__':
    unittest.main()



import os
import argparse
import warnings
import time
from PIL import Image

from .constants import ScreenType, Direction, DESCRIPTION, DEFAULTS, SUPPORTED_EXTENSIONS
from . import helpers
from . import filters


# ---- MAIN CLASSES AND FUNCTIONS ----

# The main reusable class that safely bridges user interaction with the filters
class Pixelgreat:
    def __init__(self,
                 output_size,
                 pixel_size,
                 screen_type=None,  # Set to a static default
                 direction=None,  # Set default based on screen type
                 pixel_aspect=None,  # Set to a static default
                 pixelate=None,  # Set to a static default
                 brighten=None,  # Set to a static default
                 blur=None,  # Set default based on screen type
                 washout=None,  # Set default based on screen type
                 scanline_strength=None,  # Set default based on screen type
                 scanline_spacing=None,  # Set to a static default
                 scanline_size=None,  # Set to a static default
                 scanline_blur=None,  # Set to a static default
                 grid_strength=None,  # Set to a static default
                 pixel_padding=None,  # Set default based on screen type
                 rounding=None,  # Set default based on screen type
                 bloom_strength=None,  # Set to a static default
                 bloom_size=None,  # Set to a static default
                 color_mode=None  # Set to a static default
                 ):
        # Get basic settings used for all filters
        helpers.assert_value_in_range(
            output_size[0],
            minimum=3,
            message="Output width must be no less than {min} (got {val})"
        )
        helpers.assert_value_in_range(
            output_size[1],
            minimum=3,
            message="Output height must be no less than {min} (got {val})"
        )
        self.output_size = output_size

        helpers.assert_value_in_range(
            pixel_size,
            minimum=3,
            message="Pixel size must be no less than {min} (got {val})"
        )
        if pixel_size < 10:
            warnings.warn("Pixel sizes below 10 may cause visual glitches. "
                          "Instead of making the pixels smaller, try making the output size larger.")
        self.pixel_size = pixel_size

        if screen_type is None:
            screen_type = DEFAULTS["screen_type"]
        if not isinstance(screen_type, ScreenType):
            raise ValueError("The screen_type argument must be a valid ScreenType instance")
        self.screen_type = screen_type

        if pixel_padding is None:
            # Set default based on screen type
            if self.screen_type == ScreenType.LCD:
                self.pixel_padding = DEFAULTS["pixel_padding"]["LCD"]
            elif self.screen_type == ScreenType.CRT_TV:
                self.pixel_padding = DEFAULTS["pixel_padding"]["CRT_TV"]
            else:  # Defaults to CRT monitor
                self.pixel_padding = DEFAULTS["pixel_padding"]["CRT_MONITOR"]
        else:
            helpers.assert_value_in_range(
                pixel_padding,
                minimum=0,
                maximum=1,
                message="Pixel padding must be between {min} and {max} (got {val})"
            )
            self.pixel_padding = pixel_padding

        if direction is None:
            # Set default based on screen type
            if self.screen_type == ScreenType.LCD:
                self.direction = DEFAULTS["direction"]["LCD"]
            elif self.screen_type == ScreenType.CRT_TV:
                self.direction = DEFAULTS["direction"]["CRT_TV"]
            else:  # Defaults to CRT monitor
                self.direction = DEFAULTS["direction"]["CRT_MONITOR"]
        else:
            self.direction = direction

        if washout is None:
            # Set default based on screen type
            if self.screen_type == ScreenType.CRT_TV:
                self.washout = DEFAULTS["washout"]["CRT_TV"]
            elif self.screen_type == ScreenType.CRT_MONITOR:
                self.washout = DEFAULTS["washout"]["CRT_MONITOR"]
            else:  # Defaults to LCD
                self.washout = DEFAULTS["washout"]["LCD"]
        else:
            helpers.assert_value_in_range(
                washout,
                minimum=0,
                maximum=1,
                message="Washout must be between {min} and {max} (got {val})"
            )
            self.washout = washout

        if blur is None:
            # Set default based on screen type
            if self.screen_type == ScreenType.CRT_TV:
                self.blur = DEFAULTS["blur"]["CRT_TV"]
            elif self.screen_type == ScreenType.CRT_MONITOR:
                self.blur = DEFAULTS["blur"]["CRT_MONITOR"]
            else:  # Defaults to LCD
                self.blur = DEFAULTS["blur"]["LCD"]
        else:
            helpers.assert_value_in_range(
                blur,
                minimum=0,
                maximum=1,
                message="Blur amount must be between {min} and {max} (got {val})"
            )
            self.blur = blur

        if brighten is None:
            self.brighten = DEFAULTS["brighten"]
        else:
            helpers.assert_value_in_range(
                brighten,
                minimum=0,
                maximum=1,
                message="Brighten amount must be between {min} and {max} (got {val})"
            )
            self.brighten = brighten

        if bloom_size is None:
            bloom_size = DEFAULTS["bloom_size"]
        helpers.assert_value_in_range(
            bloom_size,
            minimum=0,
            maximum=1,
            message="Bloom size must be between {min} and {max} (got {val})"
        )
        self.bloom_size = bloom_size

        if pixel_aspect is None:
            pixel_aspect = DEFAULTS["pixel_aspect"]
        helpers.assert_value_in_range(
            pixel_aspect,
            minimum=0.33,
            maximum=3,
            message="Pixel aspect ratio must be between {min} and {max} (got {val})"
        )
        self.pixel_aspect = pixel_aspect

        if rounding is None:
            # Set default based on screen type
            if self.screen_type == ScreenType.CRT_TV:
                self.rounding = DEFAULTS["rounding"]["CRT_TV"]
            elif self.screen_type == ScreenType.CRT_MONITOR:
                self.rounding = DEFAULTS["rounding"]["CRT_MONITOR"]
            else:  # Defaults to LCD
                self.rounding = DEFAULTS["rounding"]["LCD"]
        else:
            helpers.assert_value_in_range(
                rounding,
                minimum=0,
                maximum=1,
                message="Rounding must be between {min} and {max} (got {val})"
            )
            self.rounding = rounding

        if scanline_spacing is None:
            scanline_spacing = DEFAULTS["scanline_spacing"]
        helpers.assert_value_in_range(
            scanline_spacing,
            minimum=0.33,
            maximum=3,
            message="Scanline spacing must be between {min} and {max} (got {val})"
        )
        self.scanline_spacing = scanline_spacing

        if scanline_size is None:
            scanline_size = DEFAULTS["scanline_size"]
        helpers.assert_value_in_range(
            scanline_size,
            minimum=0,
            maximum=1,
            message="Scanline size must be between {min} and {max} (got {val})"
        )
        self.scanline_size = scanline_size

        if scanline_blur is None:
            scanline_blur = DEFAULTS["scanline_blur"]
        helpers.assert_value_in_range(
            scanline_blur,
            minimum=0,
            maximum=1,
            message="Scanline blur must be between {min} and {max} (got {val})"
        )
        self.scanline_blur = scanline_blur

        if scanline_strength is None:
            # Set default based on screen type
            if self.screen_type == ScreenType.CRT_TV:
                self.scanline_strength = DEFAULTS["scanline_strength"]["CRT_TV"]
            elif self.screen_type == ScreenType.CRT_MONITOR:
                self.scanline_strength = DEFAULTS["scanline_strength"]["CRT_MONITOR"]
            else:  # Defaults to LCD
                self.scanline_strength = DEFAULTS["scanline_strength"]["LCD"]
        else:
            helpers.assert_value_in_range(
                scanline_strength,
                minimum=0,
                maximum=1,
                message="Scanline strength must be between {min} and {max} (got {val})"
            )
            self.scanline_strength = scanline_strength

        if bloom_strength is None:
            bloom_strength = DEFAULTS["bloom_strength"]
        helpers.assert_value_in_range(
            bloom_strength,
            minimum=0,
            maximum=1,
            message="Bloom strength must be between {min} and {max} (got {val})"
        )
        self.bloom_strength = bloom_strength

        if grid_strength is None:
            grid_strength = DEFAULTS["grid_strength"]
        helpers.assert_value_in_range(
            grid_strength,
            minimum=0,
            maximum=1,
            message="Grid strength must be between {min} and {max} (got {val})"
        )
        self.grid_strength = grid_strength

        if pixelate is None:
            pixelate = DEFAULTS["pixelate"]
        if not isinstance(pixelate, bool):
            raise ValueError("The pixelate argument must be a valid boolean value")
        self.pixelate = pixelate

        if color_mode is None:
            color_mode = "RGB"
        color_mode = color_mode.upper()
        if "R" in color_mode and "G" in color_mode and "B" in color_mode:
            self.color_mode = color_mode
        else:
            raise ValueError(f"The color mode must have a red channel, "
                             f"a green channel, and a blue channel (got {color_mode})")

        # Compute actual pixel width based on the desired size of the smallest side
        if pixel_aspect > 1:
            # Wider than tall, use as height (to compute width)
            self.pixel_width = round(self.pixel_size * self.pixel_aspect)
        else:
            # Taller than wide, use as width
            self.pixel_width = round(self.pixel_size)

        # Create the composite filter object with the selected settings
        self.filter = filters.CompositeFilter(
            screen_type=self.screen_type,
            pixel_width=self.pixel_width,
            pixel_padding=self.pixel_padding,
            direction=self.direction,
            brighten=self.brighten,
            washout=self.washout,
            blur=self.blur,
            bloom_size=self.bloom_size,
            pixel_aspect=self.pixel_aspect,
            rounding=self.rounding,
            scanline_spacing=self.scanline_spacing,
            scanline_size=self.scanline_size,
            scanline_blur=self.scanline_blur,
            scanline_strength=self.scanline_strength,
            bloom_strength=self.bloom_strength,
            grid_strength=self.grid_strength,
            pixelate=self.pixelate,
            output_size=self.output_size,
            color_mode=self.color_mode
        )

    def apply(self, image):
        return self.filter.apply(image)

    def get_grid_filter(self, adjusted=False):
        return self.filter.get_grid_filter(adjusted=adjusted)

    def get_grid_filter_tile(self):
        return self.filter.get_grid_filter_tile()

    def get_scanline_filter(self, adjusted=False):
        return self.filter.get_scanline_filter(adjusted=adjusted)


# A single use helper function to process a single image
def pixelgreat(image,
               pixel_size,
               output_scale=None,
               screen_type=None,
               direction=None,
               pixel_aspect=None,
               pixelate=None,
               brighten=None,
               blur=None,
               washout=None,
               scanline_strength=None,
               scanline_spacing=None,
               scanline_size=None,
               scanline_blur=None,
               grid_strength=None,
               pixel_padding=None,
               rounding=None,
               bloom_strength=None,
               bloom_size=None
               ):
    if output_scale is None:
        output_scale = DEFAULTS["output_scale"]

    output_size = (
        max(round(image.width * output_scale), 3),
        max(round(image.height * output_scale), 3)
    )
    pg_object = Pixelgreat(
        output_size=output_size,
        pixel_size=pixel_size,
        screen_type=screen_type,
        pixel_padding=pixel_padding,
        direction=direction,
        washout=washout,
        brighten=brighten,
        blur=blur,
        bloom_size=bloom_size,
        pixel_aspect=pixel_aspect,
        rounding=rounding,
        scanline_spacing=scanline_spacing,
        scanline_size=scanline_size,
        scanline_blur=scanline_blur,
        scanline_strength=scanline_strength,
        bloom_strength=bloom_strength,
        grid_strength=grid_strength,
        pixelate=pixelate,
        color_mode=image.mode
    )
    result = pg_object.apply(image)

    return result


# ---- PROGRAM EXECUTION ----


# Parse arguments for a single image
def parse_args_single():
    parser = argparse.ArgumentParser(
        description=f"{DESCRIPTION}\n\n"
                    f"Valid values are shown in {{braces}}\n"
                    f"Default values are shown in [brackets]",
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    parser.add_argument("-i", "--input", dest="image_in", type=helpers.file_path, required=True,
                        help="the image to convert"
                        )

    parser.add_argument("-o", "--output", dest="image_out", type=str, required=True,
                        help="where to save the converted image, and what filetype to save it as"
                        )

    parser.add_argument("-s", "--size", dest="pixel_size", type=float, required=True,
                        help="the size of the pixels {3 - no limit}"
                        )

    parser.add_argument("-os", "--output-scale", dest="output_scale", type=float, required=False,
                        default=None,
                        help="How much to scale the output size by {no limits, 1.0 is no scaling, 2.0 is 2x size} [1.0]"
                        )

    parser.add_argument("-t", "--type", dest="screen_type", type=str, required=False,
                        default=None,
                        help="the type of RGB filter to apply {{{lcd}, {crt_tv}, {crt_mon}}} [{default}]".format(
                            lcd=ScreenType.LCD.value,
                            crt_tv=ScreenType.CRT_TV.value,
                            crt_mon=ScreenType.CRT_MONITOR.value,
                            default=DEFAULTS["screen_type"].value)
                        )

    parser.add_argument("-d", "--direction", dest="direction", type=str, required=False,
                        default=None,
                        help="the direction of the RGB filter {{{vert}, {horiz}}} [varies w/ screen type]".format(
                            vert=Direction.VERTICAL.value,
                            horiz=Direction.HORIZONTAL.value)
                        )

    parser.add_argument("-a", "--aspect", dest="pixel_aspect", type=float, required=False,
                        default=None,
                        help="the aspect ratio of the pixels, width / height {{0.33 - 3.0}} [{default}]".format(
                            default=DEFAULTS["pixel_aspect"])
                        )

    parser.add_argument("-npx", "--no-pixelate", dest="pixelate", action="store_false",
                        help="if given, the image will not be pixelated, but the other filters will still be applied"
                        )

    parser.add_argument("-br", "--brighten", dest="brighten", type=float, required=False,
                        default=None,
                        help="how much to brighten the source image {{0.0 - 1.0}} [{default}]".format(
                            default=DEFAULTS["brighten"])
                        )

    parser.add_argument("-b", "--blur", dest="blur_amount", type=float, required=False,
                        default=None,
                        help="how much to blur the source image {0.0 - 1.0} [varies w/ screen type]"
                        )

    parser.add_argument("-w", "--washout", dest="washout", type=float, required=False,
                        default=None,
                        help="how much to brighten dark pixels {0.0 - 1.0} [varies w/ screen type]"
                        )

    parser.add_argument("-sst", "--scanline-strength", dest="scanline_strength", type=float, required=False,
                        default=None,
                        help="the strength of the CRT scanline filter {0.0 - 1.0} [varies w/ screen type]"
                        )

    parser.add_argument("-ssp", "--scanline-spacing", dest="scanline_spacing", type=float, required=False,
                        default=None,
                        help="how far apart to space the CRT scanlines {{0.33 - 3.0}} [{default}]".format(
                            default=DEFAULTS["scanline_spacing"])
                        )

    parser.add_argument("-ssz", "--scanline-size", dest="scanline_size", type=float, required=False,
                        default=None,
                        help="how wide the CRT scanlines are {{0.0 - 1.0}} [{default}]".format(
                            default=DEFAULTS["scanline_size"])
                        )

    parser.add_argument("-sb", "--scanline-blur", dest="scanline_blur", type=float, required=False,
                        default=None,
                        help="how much blur to apply to the CRT scanline filter {{0.0 - 1.0}} [{default}]".format(
                            default=DEFAULTS["scanline_blur"])
                        )

    parser.add_argument("-gst", "--grid-strength", dest="grid_strength", type=float, required=False,
                        default=None,
                        help="the strength of the RGB pixel grid filter {{0.0 - 1.0}} [{default}]".format(
                            default=DEFAULTS["grid_strength"])
                        )

    parser.add_argument("-p", "--padding", dest="padding", type=float, required=False,
                        default=None,
                        help="how much black padding to add around the pixels {0.0 - 1.0} [varies w/ screen type]"
                        )

    parser.add_argument("-r", "--rounding", dest="rounding", type=float, required=False,
                        default=None,
                        help="how much to round the corners of the pixels {0.0 - 1.0} [varies w/ screen type]"
                        )

    parser.add_argument("-bst", "--bloom-strength", dest="bloom_strength", type=float, required=False,
                        default=None,
                        help="the amount of bloom to add to the output image {{0.0 - 1.0}} [{default}]".format(
                            default=DEFAULTS["bloom_strength"])
                        )

    parser.add_argument("-bsz", "--bloom-size", dest="bloom_size", type=float, required=False,
                        default=None,
                        help="the size of the bloom added to the output image {{0.0 - 1.0}} [{default}]".format(
                            default=DEFAULTS["bloom_size"])
                        )

    parsed_args = parser.parse_args()

    # Interpret string arguments
    if parsed_args.screen_type is not None:
        if parsed_args.screen_type == ScreenType.LCD.value:
            parsed_args.screen_type = ScreenType.LCD
        elif parsed_args.screen_type == ScreenType.CRT_TV.value:
            parsed_args.screen_type = ScreenType.CRT_TV
        elif parsed_args.screen_type == ScreenType.CRT_MONITOR.value:
            parsed_args.screen_type = ScreenType.CRT_MONITOR
        else:
            parser.error(f"\"{parsed_args.screen_type}\" is not a valid screen type")

    if parsed_args.direction is not None:
        if parsed_args.direction == Direction.VERTICAL.value:
            parsed_args.direction = Direction.VERTICAL
        elif parsed_args.direction == Direction.HORIZONTAL.value:
            parsed_args.direction = Direction.HORIZONTAL
        else:
            parser.error(f"\"{parsed_args.direction}\" is not a valid direction")

    # Verify the target file extensions are supported
    input_name, input_ext = os.path.splitext(parsed_args.image_in)
    input_ext = input_ext.lower()
    if input_ext not in SUPPORTED_EXTENSIONS:
        parser.error(f"\"{input_ext}\" is not a supported input format")

    output_name, output_ext = os.path.splitext(parsed_args.image_out)
    output_ext = output_ext.lower()
    if output_ext not in SUPPORTED_EXTENSIONS:
        parser.error(f"\"{output_ext}\" is not a supported output format")

    return parsed_args


# Process a single image
def single():
    # Parse args
    args = parse_args_single()

    # Open source image
    image = Image.open(os.path.realpath(args.image_in))

    start_time = time.time()

    # Apply the filter to a single image
    print("Converting image...")
    result = pixelgreat(image=image,
                        pixel_size=args.pixel_size,
                        screen_type=args.screen_type,
                        pixel_padding=args.padding,
                        direction=args.direction,
                        washout=args.washout,
                        brighten=args.brighten,
                        blur=args.blur_amount,
                        bloom_size=args.bloom_size,
                        pixel_aspect=args.pixel_aspect,
                        rounding=args.rounding,
                        scanline_spacing=args.scanline_spacing,
                        scanline_size=args.scanline_size,
                        scanline_blur=args.scanline_blur,
                        scanline_strength=args.scanline_strength,
                        bloom_strength=args.bloom_strength,
                        grid_strength=args.grid_strength,
                        pixelate=args.pixelate,
                        output_scale=args.output_scale
                        )

    # Save it
    print("Saving image...")
    output_name = os.path.realpath(args.image_out)
    output_dir = os.path.dirname(output_name)
    os.makedirs(output_dir, exist_ok=True)
    result.save(output_name)

    end_time = time.time()
    process_time = round(end_time - start_time, 1)

    print(f"Done converting 1 image in {process_time} seconds!\nSaved image: {args.image_out}")


# Parse arguments for an image sequence
def parse_args_sequence():
    parser = argparse.ArgumentParser(
        description=f"{DESCRIPTION}\n\n"
                    f"Valid values are shown in {{braces}}\n"
                    f"Default values are shown in [brackets]\n\n"
                    f"For image sequences, the output size is based on the first image in the sequence",
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    parser.add_argument("-i", "--input", dest="image_in", type=helpers.file_path, required=True,
                        help="the image to convert (must be part of a sequence)"
                        )

    parser.add_argument("-o", "--output", dest="image_out", type=str, required=True,
                        help="where to save the converted image sequence, and what filetype to save them as"
                        )

    parser.add_argument("-s", "--size", dest="pixel_size", type=float, required=True,
                        help="the size of the pixels {3 - no limit}"
                        )

    parser.add_argument("-os", "--output-scale", dest="output_scale", type=float, required=False,
                        default=None,
                        help="How much to scale the output size by {no limits, 1.0 is no scaling, 2.0 is 2x size} [1.0]"
                        )

    parser.add_argument("-t", "--type", dest="screen_type", type=str, required=False,
                        default=None,
                        help="the type of RGB filter to apply {{{lcd}, {crt_tv}, {crt_mon}}} [{default}]".format(
                            lcd=ScreenType.LCD.value,
                            crt_tv=ScreenType.CRT_TV.value,
                            crt_mon=ScreenType.CRT_MONITOR.value,
                            default=DEFAULTS["screen_type"].value)
                        )

    parser.add_argument("-d", "--direction", dest="direction", type=str, required=False,
                        default=None,
                        help="the direction of the RGB filter {{{vert}, {horiz}}} [varies w/ screen type]".format(
                            vert=Direction.VERTICAL.value,
                            horiz=Direction.HORIZONTAL.value)
                        )

    parser.add_argument("-a", "--aspect", dest="pixel_aspect", type=float, required=False,
                        default=None,
                        help="the aspect ratio of the pixels, width / height {{0.33 - 3.0}} [{default}]".format(
                            default=DEFAULTS["pixel_aspect"])
                        )

    parser.add_argument("-npx", "--no-pixelate", dest="pixelate", action="store_false",
                        help="if given, the image will not be pixelated, but the other filters will still be applied"
                        )

    parser.add_argument("-br", "--brighten", dest="brighten", type=float, required=False,
                        default=None,
                        help="how much to brighten the source image {{0.0 - 1.0}} [{default}]".format(
                            default=DEFAULTS["brighten"])
                        )

    parser.add_argument("-b", "--blur", dest="blur_amount", type=float, required=False,
                        default=None,
                        help="how much to blur the source image {0.0 - 1.0} [varies w/ screen type]"
                        )

    parser.add_argument("-w", "--washout", dest="washout", type=float, required=False,
                        default=None,
                        help="how much to brighten dark pixels {0.0 - 1.0} [varies w/ screen type]"
                        )

    parser.add_argument("-sst", "--scanline-strength", dest="scanline_strength", type=float, required=False,
                        default=None,
                        help="the strength of the CRT scanline filter {0.0 - 1.0} [varies w/ screen type]"
                        )

    parser.add_argument("-ssp", "--scanline-spacing", dest="scanline_spacing", type=float, required=False,
                        default=None,
                        help="how far apart to space the CRT scanlines {{0.33 - 3.0}} [{default}]".format(
                            default=DEFAULTS["scanline_spacing"])
                        )

    parser.add_argument("-ssz", "--scanline-size", dest="scanline_size", type=float, required=False,
                        default=None,
                        help="how wide the CRT scanlines are {{0.0 - 1.0}} [{default}]".format(
                            default=DEFAULTS["scanline_size"])
                        )

    parser.add_argument("-sb", "--scanline-blur", dest="scanline_blur", type=float, required=False,
                        default=None,
                        help="how much blur to apply to the CRT scanline filter {{0.0 - 1.0}} [{default}]".format(
                            default=DEFAULTS["scanline_blur"])
                        )

    parser.add_argument("-gst", "--grid-strength", dest="grid_strength", type=float, required=False,
                        default=None,
                        help="the strength of the RGB pixel grid filter {{0.0 - 1.0}} [{default}]".format(
                            default=DEFAULTS["grid_strength"])
                        )

    parser.add_argument("-p", "--padding", dest="padding", type=float, required=False,
                        default=None,
                        help="how much black padding to add around the pixels {0.0 - 1.0} [varies w/ screen type]"
                        )

    parser.add_argument("-r", "--rounding", dest="rounding", type=float, required=False,
                        default=None,
                        help="how much to round the corners of the pixels {0.0 - 1.0} [varies w/ screen type]"
                        )

    parser.add_argument("-bst", "--bloom-strength", dest="bloom_strength", type=float, required=False,
                        default=None,
                        help="the amount of bloom to add to the output image {{0.0 - 1.0}} [{default}]".format(
                            default=DEFAULTS["bloom_strength"])
                        )

    parser.add_argument("-bsz", "--bloom-size", dest="bloom_size", type=float, required=False,
                        default=None,
                        help="the size of the bloom added to the output image {{0.0 - 1.0}} [{default}]".format(
                            default=DEFAULTS["bloom_size"])
                        )

    parsed_args = parser.parse_args()

    # Interpret string arguments
    if parsed_args.screen_type is not None:
        if parsed_args.screen_type == ScreenType.LCD.value:
            parsed_args.screen_type = ScreenType.LCD
        elif parsed_args.screen_type == ScreenType.CRT_TV.value:
            parsed_args.screen_type = ScreenType.CRT_TV
        elif parsed_args.screen_type == ScreenType.CRT_MONITOR.value:
            parsed_args.screen_type = ScreenType.CRT_MONITOR
        else:
            parser.error(f"\"{parsed_args.screen_type}\" is not a valid screen type")

    if parsed_args.direction is not None:
        if parsed_args.direction == Direction.VERTICAL.value:
            parsed_args.direction = Direction.VERTICAL
        elif parsed_args.direction == Direction.HORIZONTAL.value:
            parsed_args.direction = Direction.HORIZONTAL
        else:
            parser.error(f"\"{parsed_args.direction}\" is not a valid direction")

    # Verify the target file extensions are supported
    input_name, input_ext = os.path.splitext(parsed_args.image_in)
    input_ext = input_ext.lower()
    if input_ext not in SUPPORTED_EXTENSIONS:
        parser.error(f"\"{input_ext}\" is not a supported input format")

    output_name, output_ext = os.path.splitext(parsed_args.image_out)
    output_ext = output_ext.lower()
    if output_ext not in SUPPORTED_EXTENSIONS:
        parser.error(f"\"{output_ext}\" is not a supported output format")

    # Validate the input image actually represents an image sequence
    if helpers.parse_sequenced_image_name(parsed_args.image_in)["error"] is not None:
        parser.error("No image sequence found. Ensure they are named like this: name0000.png, name0001.png, etc.")

    # Set default scale
    if parsed_args.output_scale is None:
        parsed_args.output_scale = 1.0

    return parsed_args


# Process an image sequence
def sequence():
    args = parse_args_sequence()
    print("Preparing to process image sequence...")

    start_time = time.time()

    # Get the full image sequence
    sequence_info = helpers.get_all_images_in_sequence(args.image_in)

    # Get the size and color mode of the first image
    first_image = Image.open(sequence_info["files"][0])
    first_image_size = first_image.size
    first_image_mode = first_image.mode
    first_image.close()

    # Get the output size
    output_size = (
        max(round(first_image_size[0] * args.output_scale), 3),
        max(round(first_image_size[1] * args.output_scale), 3)
    )

    # Make the re-usable converter object
    converter = Pixelgreat(
        output_size=output_size,
        pixel_size=args.pixel_size,
        screen_type=args.screen_type,
        pixel_padding=args.padding,
        direction=args.direction,
        washout=args.washout,
        brighten=args.brighten,
        blur=args.blur_amount,
        bloom_size=args.bloom_size,
        pixel_aspect=args.pixel_aspect,
        rounding=args.rounding,
        scanline_spacing=args.scanline_spacing,
        scanline_size=args.scanline_size,
        scanline_blur=args.scanline_blur,
        scanline_strength=args.scanline_strength,
        bloom_strength=args.bloom_strength,
        grid_strength=args.grid_strength,
        pixelate=args.pixelate,
        color_mode=first_image_mode
    )

    # Make the destination dir if it doesn't already exist
    output_dir = os.path.dirname(args.image_out)
    os.makedirs(output_dir, exist_ok=True)

    # Loop through the images
    image_count = len(sequence_info["files"])
    for i, image_name in enumerate(sequence_info["files"]):
        print(f"Converting image {i + 1} of {image_count}...")

        # Open image
        image_in = Image.open(image_name)

        # Convert the image with the reusable converter
        image_out = converter.apply(image_in)

        # Get new image filename
        main_name, ext = os.path.splitext(args.image_out)
        this_number = str(i).rjust(sequence_info["digits"], "0")
        output_name = f"{main_name}{this_number}{ext}"

        print(f"  Saving image...")

        # Save the image
        try:
            image_out.save(output_name)
        except OSError as e:
            # Try converting it to RGB first
            rgb_image_out = image_out.convert("RGB")
            rgb_image_out.save(output_name)
            rgb_image_out.close()

        # Close the images
        image_in.close()
        image_out.close()

    end_time = time.time()
    process_time = round(end_time - start_time, 1)

    print(f"Done converting {image_count} images in {process_time} seconds!\n"
          f"Saved images to {output_dir}")



import math
from PIL import Image, ImageDraw, ImageChops, ImageFilter, ImageEnhance

from . import helpers
from .constants import Direction, ScreenType

# TODO: XO-1 LCD Display


def lcd(pixel_width, padding, direction, aspect, rounding, color_mode="RGB", subpixels=8):
    # Get main pixel dimensions (float, used in calculations)
    if direction == Direction.HORIZONTAL:
        # Adjust aspect for later rotation if needed
        aspect = 1 / aspect

        real_width = pixel_width * aspect
        real_height = pixel_width
    else:
        real_width = pixel_width
        real_height = pixel_width / aspect

    # Get width and height based on subpixels
    if subpixels > 1:
        width = real_width * subpixels
        height = real_height * subpixels
    else:
        width = real_width
        height = real_height

    # Get integer values for width and height (the REAL pixel count)
    width_px = round(width)
    height_px = round(height)

    # Make new black image
    filter_image = Image.new(color_mode, (width_px, height_px), (0, 0, 0))

    # Make drawing object
    filter_draw = ImageDraw.Draw(filter_image)

    # Get padding pixels based on size (1.0 should just barely block out all RGB, 0.0 should equal 0 px)
    if aspect < 3.0:
        padding_px = round(padding * (width / 3))
    else:
        padding_px = round(padding * height)

    # If padding is 0, make sure it's actually 0 so that we can draw full color bars later
    if padding <= 0.0:
        padding_px = 0

    # If it's 1.0, we know it should be full black, so we should return the black image now
    if padding >= 1.0:
        return filter_image

    # Get left and right padding amounts
    padding_left = math.ceil(padding_px / 2)  # Also used for tops
    padding_right = padding_px - padding_left  # Also used for bottoms

    # Get variables use for drawing the color regions
    rg_boundary_x = round(width / 3)
    gb_boundary_x = round((2 * width) / 3)

    if padding_px == 0:
        red_start = (0, 0)
        red_end = (rg_boundary_x - 1, height_px - 1)

        green_start = (rg_boundary_x, 0)
        green_end = (gb_boundary_x - 1, height_px - 1)

        blue_start = (gb_boundary_x, 0)
        blue_end = (width_px - 1, height_px - 1)
    else:
        red_start = (padding_left, padding_left)
        red_end = (
            max((rg_boundary_x - padding_right) - 1, red_start[0]),
            max((height_px - padding_right) - 1, red_start[1])
        )

        green_start = (rg_boundary_x + padding_left, padding_left)
        green_end = (
            max((gb_boundary_x - padding_right) - 1, green_start[0]),
            max((height_px - padding_right) - 1, green_start[1])
        )

        blue_start = (gb_boundary_x + padding_left, padding_left)
        blue_end = (
            max((width_px - padding_right) - 1, blue_start[0]),
            max((height_px - padding_right) - 1, blue_start[1])
        )

    color_size_px = red_end[0] - red_start[0]

    # Draw color regions
    if rounding > 0:
        round_radius = round((color_size_px / 2) * rounding)

        filter_draw.rounded_rectangle((red_start, red_end), fill=(255, 0, 0), radius=round_radius)
        filter_draw.rounded_rectangle((green_start, green_end), fill=(0, 255, 0), radius=round_radius)
        filter_draw.rounded_rectangle((blue_start, blue_end), fill=(0, 0, 255), radius=round_radius)
    else:
        filter_draw.rectangle((red_start, red_end), fill=(255, 0, 0))
        filter_draw.rectangle((green_start, green_end), fill=(0, 255, 0))
        filter_draw.rectangle((blue_start, blue_end), fill=(0, 0, 255))

    # Scale down if needed
    if subpixels > 1:
        filter_image = filter_image.resize(
            (
                round(real_width),
                round(real_width)
            ),
            resample=Image.Resampling.LANCZOS
        )

    # Rotate if needed
    if direction == Direction.HORIZONTAL:
        filter_image = filter_image.rotate(270, expand=True)

    return filter_image


def crt_tv(pixel_width, padding, direction, aspect, rounding, color_mode="RGB", subpixels=8):
    # Get the first half of the filter
    single_pixel = lcd(pixel_width=pixel_width,
                       padding=padding,
                       direction=direction,
                       aspect=aspect,
                       rounding=rounding,
                       color_mode=color_mode,
                       subpixels=subpixels
                       )

    # Figure out the dimensions for the new filter
    if direction == Direction.VERTICAL:
        new_size = (single_pixel.width * 2, single_pixel.height)
    else:
        new_size = (single_pixel.width, single_pixel.height * 2)

    # Make new image to paste old one onto
    both_pixels = Image.new(color_mode, new_size, color=(0, 0, 0))

    # Paste the first pixel
    both_pixels.paste(single_pixel, (0, 0))

    # Paste the second pixel in two parts
    if direction == Direction.VERTICAL:
        split_y = round(both_pixels.height / 2)
        both_pixels.paste(single_pixel, (single_pixel.width, split_y))
        both_pixels.paste(single_pixel, (single_pixel.width, split_y - single_pixel.height))
    else:
        split_x = round(both_pixels.width / 2)
        both_pixels.paste(single_pixel, (split_x, single_pixel.height))
        both_pixels.paste(single_pixel, (split_x - single_pixel.width, single_pixel.height))

    return both_pixels


def crt_monitor(pixel_width, padding, direction, color_mode="RGB", subpixels=8):
    # Adjust size to more correctly match real mapping for pixel sizes
    pixel_width = (pixel_width / math.sqrt(3))

    # Get width and height from dot size (float, used for calculations)
    real_width = pixel_width * 3
    real_height = pixel_width * math.sqrt(3)

    # Get size, width, and height based on subpixels
    if subpixels > 1:
        pixel_width = pixel_width * subpixels

        width = real_width * subpixels
        height = real_height * subpixels
    else:
        width = real_width
        height = real_height

    # Get integer divisions for dot placement
    height_divs = (
        0,
        round(height / 2),
        round(height)
    )
    width_divs = (
        0,
        round(width * (1 / 6)),
        round(width * (2 / 6)),
        round(width * (3 / 6)),
        round(width * (4 / 6)),
        round(width * (5 / 6)),
        round(width)
    )

    # Make new black image
    filter_image = Image.new(color_mode, (width_divs[-1], height_divs[-1]), (0, 0, 0))

    # Make drawing object
    filter_draw = ImageDraw.Draw(filter_image)

    # If padding is  1, know it should be full black, so we should return the black image now
    if padding == 1:
        return filter_image

    # Get dot size
    dot_size = pixel_width * (1 - padding)

    # Draw red dots
    filter_draw.ellipse(
        helpers.get_centered_dimensions((width_divs[0], height_divs[0]), (dot_size, dot_size)),
        fill=(255, 0, 0)
    )
    filter_draw.ellipse(
        helpers.get_centered_dimensions((width_divs[6], height_divs[0]), (dot_size, dot_size)),
        fill=(255, 0, 0)
    )
    filter_draw.ellipse(
        helpers.get_centered_dimensions((0, height_divs[2]), (dot_size, dot_size)),
        fill=(255, 0, 0)
    )
    filter_draw.ellipse(
        helpers.get_centered_dimensions((width_divs[6], height_divs[2]), (dot_size, dot_size)),
        fill=(255, 0, 0)
    )
    filter_draw.ellipse(
        helpers.get_centered_dimensions((width_divs[3], height_divs[1]), (dot_size, dot_size)),
        fill=(255, 0, 0)
    )

    # Draw green dots
    filter_draw.ellipse(
        helpers.get_centered_dimensions((width_divs[2], height_divs[0]), (dot_size, dot_size)),
        fill=(0, 255, 0)
    )
    filter_draw.ellipse(
        helpers.get_centered_dimensions((width_divs[2], height_divs[2]), (dot_size, dot_size)),
        fill=(0, 255, 0)
    )
    filter_draw.ellipse(
        helpers.get_centered_dimensions((width_divs[5], height_divs[1]), (dot_size, dot_size)),
        fill=(0, 255, 0)
    )

    # Draw blue dots
    filter_draw.ellipse(
        helpers.get_centered_dimensions((width_divs[1], height_divs[1]), (dot_size, dot_size)),
        fill=(0, 0, 255)
    )
    filter_draw.ellipse(
        helpers.get_centered_dimensions((width_divs[4], height_divs[0]), (dot_size, dot_size)),
        fill=(0, 0, 255)
    )
    filter_draw.ellipse(
        helpers.get_centered_dimensions((width_divs[4], height_divs[2]), (dot_size, dot_size)),
        fill=(0, 0, 255)
    )

    # Scale down if needed
    if subpixels > 1:
        filter_image = filter_image.resize(
            (
                round(real_width),
                round(real_height)
            ),
            resample=Image.Resampling.LANCZOS
        )

    # Rotate if needed
    if direction == Direction.VERTICAL:
        filter_image = filter_image.rotate(270, expand=True)

    return filter_image


def scanlines(size, spacing, offset, line_size, blur, direction, color_mode="RGB"):
    # Create new black image for building the scanline filter
    scanline_image = Image.new(color_mode, size, color=(0, 0, 0))

    # If the line size is 0, we know it should be all black
    if line_size == 0:
        return scanline_image

    # Make drawing object
    scanline_draw = ImageDraw.Draw(scanline_image)

    # Get line width (integer)
    line_width = round(spacing * line_size)

    # Draw hard lines
    if direction == Direction.HORIZONTAL:
        line_count = math.ceil(scanline_image.height / spacing)
        for line in range(line_count):
            line_start_y = round((line * spacing) + (offset * spacing))
            line_end_y = line_start_y + line_width

            scanline_draw.rectangle(((0, line_start_y), (scanline_image.width, line_end_y)), fill=(255, 255, 255))
    else:
        line_count = math.ceil(scanline_image.width / spacing)
        for line in range(line_count):
            line_start_x = round((line * spacing) + (offset * spacing))
            line_end_x = line_start_x + line_width

            scanline_draw.rectangle(((line_start_x, 0), (line_end_x, scanline_image.height)), fill=(255, 255, 255))

    # Apply blur
    if blur > 0:
        blur_amt = line_width * blur
        scanline_image = scanline_image.filter(ImageFilter.GaussianBlur(blur_amt))

    return scanline_image


# Calculate approximate pixel count to match target pixel size
def get_approximate_pixel_count(size, pixel_width, pixel_aspect, make_int=True):
    # Calculate pixel height
    pixel_height = pixel_width / pixel_aspect

    # Calculate approximate pixel counts to match target size
    pixels_wide = size[0] / pixel_width
    pixels_tall = size[1] / pixel_height

    # Round if needed
    if make_int:
        pixels_wide = round(pixels_wide)
        pixels_tall = round(pixels_tall)

    return pixels_wide, pixels_tall


def pixelate_image(image,
                   pixel_width,
                   pixel_aspect,
                   output_size=None,  # Defaults to the input image size
                   downscale_mode=Image.Resampling.HAMMING
                   ):
    if output_size is None:
        output_size = image.size

    pixels_wide, pixels_tall = get_approximate_pixel_count(
        size=output_size,
        pixel_width=pixel_width,
        pixel_aspect=pixel_aspect
    )

    # Downscale image
    small_image = image.resize((pixels_wide, pixels_tall), resample=downscale_mode)

    # Upscale to the final size
    result = small_image.resize(output_size, resample=Image.Resampling.NEAREST)

    return result


def bloom_image(image, bloom_size, bloom_strength):
    if bloom_strength <= 0 or bloom_size <= 0:
        return image

    # Blur source image
    bloom = image.filter(ImageFilter.GaussianBlur(bloom_size))

    # Adjust the blurred image
    bloom = helpers.mix_color_with_image(
        bloom,
        (0, 0, 0),
        1 - bloom_strength
    )

    # Make final image
    result = ImageChops.lighter(image, bloom)

    return result


# A reusable class to handle applying scanlines
class ScanlineFilter:
    def __init__(self,
                 size,
                 line_spacing,
                 line_offset,
                 line_size,
                 line_blur,
                 direction,
                 strength=1.0,
                 color_mode="RGB"
                 ):
        self.size = size

        self.line_spacing = max(line_spacing, 1)

        self.line_offset = line_offset

        self.line_size = line_size

        self.line_blur = line_blur

        self.direction = direction

        self.strength = strength

        self.color_mode = color_mode

        # Pre-compute filter image
        self.filter_raw = scanlines(
            size=self.size,
            spacing=self.line_spacing,
            offset=self.line_offset,
            line_size=self.line_size,
            blur=self.line_blur,
            direction=self.direction,
            color_mode=self.color_mode
        )

        # Pre-computed adjusted filter image
        self.filter = helpers.mix_color_with_image(
            self.filter_raw,
            (255, 255, 255),
            1 - self.strength
        )

    # Apply the filter to a desired image
    def apply(self, image):
        if image.size != self.size:
            raise ValueError(f"Input image size \"{image.size}\" "
                             f"does not match filter size \"{self.size}\"")
        if image.mode != self.color_mode:
            raise ValueError(f"Input image color mode \"{image.mode}\" "
                             f"does not match filter color mode \"{self.color_mode}\"")

        result = ImageChops.multiply(image, self.filter)

        return result

    def get_filter(self, adjusted=False):
        return self.filter_raw


# A reusable class to handle applying the RGB filter
class ScreenFilter:
    def __init__(self,
                 size,
                 screen_type,
                 pixel_width,
                 pixel_padding,
                 direction,
                 pixel_aspect=None,
                 rounding=None,
                 strength=1.0,
                 color_mode="RGB"
                 ):
        self.size = size

        self.screen_type = screen_type

        self.pixel_width = pixel_width

        self.pixel_padding = pixel_padding

        self.direction = direction

        self.pixel_aspect = pixel_aspect
        self.rounding = rounding

        self.color_mode = color_mode

        # Get the filter tile image
        if self.screen_type == ScreenType.CRT_MONITOR:
            self.filter_tile = crt_monitor(
                pixel_width=self.pixel_width,
                padding=self.pixel_padding,
                direction=self.direction,
                color_mode=self.color_mode
            )
        elif self.screen_type == ScreenType.CRT_TV:
            self.filter_tile = crt_tv(
                pixel_width=self.pixel_width,
                padding=self.pixel_padding,
                direction=self.direction,
                aspect=self.pixel_aspect,
                rounding=self.rounding,
                color_mode=self.color_mode
            )
        else:  # Default to LCD
            self.filter_tile = lcd(
                pixel_width=self.pixel_width,
                padding=self.pixel_padding,
                direction=self.direction,
                aspect=self.pixel_aspect,
                rounding=self.rounding,
                color_mode=self.color_mode
            )

        self.strength = strength

        # Compute the actual counts to tile with based on the screen type and size vars
        if self.screen_type == ScreenType.CRT_MONITOR:
            # 3:sqrt(3) inherent ratio, fixed
            # We can just tile normally without worrying about alignment
            self.pixel_count = None
        elif self.screen_type == ScreenType.CRT_TV:
            # 2:1 inherent ratio, changes with pixel_aspect
            self.pixel_count = get_approximate_pixel_count(
                pixel_width=self.pixel_width,
                pixel_aspect=self.pixel_aspect,
                size=self.size,
                make_int=False
            )

            # Adjust the correct dim and round to nearest 0.5
            if self.direction == Direction.VERTICAL:
                self.pixel_count = (
                    helpers.round_to_division(self.pixel_count[0] / 2, 0.5),
                    round(self.pixel_count[1])
                )
            else:
                self.pixel_count = (
                    round(self.pixel_count[0]),
                    helpers.round_to_division(self.pixel_count[1] / 2, 0.5)
                )
        else:  # Default to LCD
            # 1:1 inherent ratio, changes with pixel_aspect
            self.pixel_count = get_approximate_pixel_count(
                pixel_width=self.pixel_width,
                pixel_aspect=self.pixel_aspect,
                size=self.size
            )

        # Pre-compute a tiled filter image
        self.filter_raw = helpers.tile_image(
            self.filter_tile,
            self.size,
            background_color=(0, 0, 0),
            count=self.pixel_count
        )

        # Pre-computed adjusted filter image
        self.filter = helpers.mix_color_with_image(
            self.filter_raw,
            (255, 255, 255),
            1 - self.strength
        )

    # Apply the filter to a desired image
    def apply(self, image):
        if image.size != self.size:
            raise ValueError(f"Input image size \"{image.size}\" "
                             f"does not match filter size \"{self.size}\"")
        if image.mode != self.color_mode:
            raise ValueError(f"Input image color mode \"{image.mode}\" "
                             f"does not match filter color mode \"{self.color_mode}\"")

        result = ImageChops.multiply(image, self.filter)

        return result

    def get_filter(self, adjusted=False):
        if adjusted:
            return self.filter
        else:
            return self.filter_raw

    def get_filter_tile(self):
        return self.filter_tile


# A reusable class to handle the composite effects
class CompositeFilter:
    def __init__(self,
                 output_size,
                 screen_type,
                 pixel_width,
                 pixel_padding,
                 direction,
                 brighten=1.0,
                 washout=0.0,
                 blur=0.0,
                 bloom_size=0.0,
                 pixel_aspect=None,
                 rounding=None,
                 scanline_spacing=None,
                 scanline_size=None,
                 scanline_blur=None,
                 scanline_strength=None,  # Default set based on screen_type
                 bloom_strength=1.0,
                 grid_strength=1.0,
                 pixelate=True,
                 color_mode="RGB"
                 ):
        self.screen_type = screen_type

        self.pixel_width = pixel_width

        self.pixel_padding = pixel_padding

        self.direction = direction

        self.brighten = brighten
        self.brighten_value = 1.0 + ((1/3) * self.brighten)

        self.washout = washout
        self.washout_value = round((self.washout * 255) / 10)

        self.blur = blur
        self.blur_px = round((self.pixel_width / 2) * self.blur)

        self.pixel_aspect = pixel_aspect

        self.bloom_size = bloom_size
        # Set actual size based on smallest pixel dimension
        if self.pixel_aspect < 1.0:
            # Use pixel size (width)
            self.bloom_size_px = round((self.pixel_width / 2) * self.bloom_size)
        else:
            # Use pixel height
            self.bloom_size_px = round(((self.pixel_width / self.pixel_aspect) / 2) * self.bloom_size)

        self.rounding = rounding

        self.scanline_size = scanline_size
        self.scanline_offset = -self.scanline_size / 2
        # Adjust for LCD
        if self.screen_type == ScreenType.LCD:
            self.scanline_offset += 0.5

        self.scanline_spacing = scanline_spacing
        # Set actual size based on the direction
        if self.screen_type == ScreenType.CRT_MONITOR:
            # Adjust the spacing for the hexagonal grid used in the CRT Monitor effect
            self.scanline_spacing_px = round((self.pixel_width * self.scanline_spacing) / 2)
        elif self.direction == Direction.HORIZONTAL:
            # Use pixel size (width)
            self.scanline_spacing_px = round(self.pixel_width * self.scanline_spacing)
        else:
            # Use pixel height
            self.scanline_spacing_px = round((self.pixel_width / self.pixel_aspect) * self.scanline_spacing)
        # Adjust for the CRT TV
        if self.screen_type == ScreenType.CRT_TV:
            self.scanline_spacing_px = round(self.scanline_spacing_px / 2)

        self.scanline_blur = scanline_blur

        self.scanline_strength = scanline_strength
        if self.scanline_strength is None:
            if self.screen_type in [ScreenType.CRT_TV, ScreenType.CRT_MONITOR]:
                self.scanline_strength = 1.0
            else:
                self.scanline_strength = 0.0

        self.bloom_strength = bloom_strength

        self.grid_strength = grid_strength

        self.pixelate = pixelate

        self.output_size = output_size

        self.color_mode = color_mode

        # Get scanline direction based on screen type
        if self.screen_type in [ScreenType.CRT_MONITOR]:
            self.scanline_direction = self.direction
        else:
            if self.direction == Direction.HORIZONTAL:
                self.scanline_direction = Direction.VERTICAL
            else:
                self.scanline_direction = Direction.HORIZONTAL

        # Make sure we have required variables for each situation
        if self.grid_strength > 0:
            if self.screen_type in [ScreenType.LCD, ScreenType.CRT_TV]:
                if self.pixel_aspect is None:
                    raise ValueError("This screen type requires the argument pixel_aspect")
                if self.rounding is None:
                    raise ValueError("This screen type requires the argument rounding")

        if self.scanline_strength > 0:
            if self.scanline_spacing is None:
                raise ValueError("Scanlines are enabled, requires the argument scanline_spacing")
            if self.scanline_size is None:
                raise ValueError("Scanlines are enabled, requires the argument scanline_size")
            if self.scanline_blur is None:
                raise ValueError("Scanlines are enabled, requires the argument scanline_blur")

        if self.pixelate:
            if self.pixel_aspect is None:
                raise ValueError("Pixelate enabled, requires the argument pixel_aspect")

        # Make the scanline filter object (if needed)
        if self.scanline_strength > 0:
            self.scanline_filter = ScanlineFilter(
                size=self.output_size,
                line_spacing=self.scanline_spacing_px,
                line_offset=self.scanline_offset,
                line_size=self.scanline_size,
                line_blur=self.scanline_blur,
                direction=self.scanline_direction,
                strength=self.scanline_strength,
                color_mode=self.color_mode
            )
        else:
            self.scanline_filter = None

        # Make the screen filter object (if needed)
        if self.grid_strength > 0:
            self.screen_filter = ScreenFilter(
                size=self.output_size,
                screen_type=self.screen_type,
                pixel_width=self.pixel_width,
                pixel_padding=self.pixel_padding,
                direction=self.direction,
                pixel_aspect=self.pixel_aspect,
                rounding=self.rounding,
                strength=self.grid_strength,
                color_mode=self.color_mode
            )
        else:
            self.screen_filter = None

    # Apply the filter to a desired image
    def apply(self, image):
        # Make input image the correct color mode
        if image.mode != self.color_mode:
            image = image.convert(self.color_mode)

        # Brighten if applicable
        if self.brighten > 0:
            image = ImageEnhance.Contrast(image).enhance(self.brighten_value)
            image = ImageEnhance.Brightness(image).enhance(self.brighten_value)

        # Pixelate / scale to final size
        if self.pixelate:
            result = pixelate_image(
                image=image,
                pixel_width=self.pixel_width,
                pixel_aspect=self.pixel_aspect,
                output_size=self.output_size
            )
        else:
            if image.size != self.output_size:
                result = image.resize(self.output_size, resample=image.Resampling.NEAREST)
            else:
                result = image.copy()

        # Blur, if relevant
        if self.blur > 0:
            result = result.filter(ImageFilter.GaussianBlur(self.blur_px))

        # Add scanlines if applicable
        if self.scanline_filter is not None:
            result = self.scanline_filter.apply(result)

        # Apply washout, if needed
        if self.washout > 0:
            result = ImageChops.lighter(
                result,
                Image.new(
                    self.color_mode,
                    self.output_size,
                    (self.washout_value, self.washout_value, self.washout_value)
                )
            )

        # Add pixel grid if applicable
        if self.screen_filter is not None:
            result = self.screen_filter.apply(result)

        # Add bloom, if applicable
        if self.bloom_size_px > 0 and self.bloom_strength > 0:
            result = bloom_image(result, self.bloom_size_px, self.bloom_strength)

        return result

    def get_grid_filter(self, adjusted=False):
        if self.screen_filter is not None:
            return self.screen_filter.get_filter(adjusted=adjusted)
        else:
            return None

    def get_grid_filter_tile(self):
        if self.screen_filter is not None:
            return self.screen_filter.get_filter_tile()
        else:
            return None

    def get_scanline_filter(self, adjusted=False):
        if self.scanline_filter is not None:
            return self.scanline_filter.get_filter(adjusted=adjusted)
        else:
            return None



from enum import Enum
from PIL import Image

DESCRIPTION = "A highly realistic RGB pixel filter"


class Direction(Enum):
    VERTICAL = "V"
    HORIZONTAL = "H"


class ScreenType(Enum):
    LCD = "LCD"
    CRT_TV = "CRT_TV"
    CRT_MONITOR = "CRT_MONITOR"


DEFAULTS = {
    "screen_type": ScreenType.LCD,
    "pixel_padding": {
        "LCD": 0.25,
        "CRT_TV": 0.25,
        "CRT_MONITOR": 0.1
    },
    "direction": {
        "LCD": Direction.VERTICAL,
        "CRT_TV": Direction.VERTICAL,
        "CRT_MONITOR": Direction.HORIZONTAL
    },
    "washout": {
        "LCD": 0.1,
        "CRT_TV": 0.5,
        "CRT_MONITOR": 0.5
    },
    "brighten": 1.0,
    "blur": {
        "LCD": 0,
        "CRT_TV": 0.5,
        "CRT_MONITOR": 0.75
    },
    "bloom_size": 0.5,
    "pixel_aspect": 1.0,
    "rounding": {
        "LCD": 0,
        "CRT_TV": 0.5,
        "CRT_MONITOR": 0
    },
    "scanline_spacing": 0.79,
    "scanline_size": 0.75,
    "scanline_blur": 0.25,
    "scanline_strength": {
        "LCD": 0,
        "CRT_TV": 1,
        "CRT_MONITOR": 0.5
    },
    "bloom_strength": 1.0,
    "grid_strength": 1.0,
    "pixelate": True,
    "output_scale": 1.0
}

SUPPORTED_EXTENSIONS = tuple([ex for ex, f in Image.registered_extensions().items() if f in Image.OPEN])



import unittest
import os
import shutil
from PIL import Image

from pixelgreat import filters

tests_dir = os.path.dirname(os.path.realpath(__file__))

test_image = Image.open(os.path.join(tests_dir, "images", "PM5544.png"))
save_dir = os.path.join(tests_dir, "processed_images")


class MakeManualTestAssets(unittest.TestCase):
    def make_image(self,
                   save_location,
                   screen_type,
                   pixel_size,
                   pixel_padding,
                   direction,
                   pixel_aspect,
                   washout=0.0,
                   blur=0.0,
                   bloom_size=0.0,
                   rounding=0.0,
                   scanline_spacing=1.0,
                   scanline_size=0.75,
                   scanline_blur=0.25,
                   scanline_strength=1.0,
                   bloom_strength=1.0,
                   grid_strength=1.0,
                   pixelate=True,
                   output_scale=1
                   ):

        # Compute image
        image = filters.CompositeFilter(
            size=test_image.size,
            screen_type=screen_type,
            pixel_width=pixel_size,
            pixel_padding=pixel_padding,
            direction=direction,
            washout=washout,
            blur=blur,
            bloom_size=bloom_size,
            pixel_aspect=pixel_aspect,
            rounding=rounding,
            scanline_spacing=scanline_spacing,
            scanline_size=scanline_size,
            scanline_blur=scanline_blur,
            scanline_strength=scanline_strength,
            bloom_strength=bloom_strength,
            grid_strength=grid_strength,
            pixelate=pixelate,
            output_size=(
                round(test_image.width * output_scale),
                round(test_image.height * output_scale)
            ),
            color_mode=test_image.mode
        ).apply(test_image)

        # Build filename
        output_name = ""

        if screen_type == filters.ScreenType.LCD:
            output_name += "LCD"
        elif screen_type == filters.ScreenType.CRT_TV:
            output_name += "TV"
        else:
            output_name += "MON"

        output_name += f"_pxs{pixel_size}"

        output_name += f"_pxa{pixel_aspect:.2f}"

        if direction == filters.Direction.VERTICAL:
            output_name += "_dV"
        elif direction == filters.Direction.HORIZONTAL:
            output_name += "_dH"

        if scanline_strength > 0:
            output_name += f"_ssp{scanline_spacing:.2f}"

        if grid_strength > 0:
            output_name += f"_pxp{pixel_padding:.2f}"

        if washout > 0:
            output_name += f"_w{washout:.2f}"

        if blur > 0:
            output_name += f"_b{blur:.2f}"

        if bloom_size > 0 and bloom_strength > 0:
            output_name += f"_bs{bloom_size:.2f}"

        if screen_type in [filters.ScreenType.LCD, filters.ScreenType.CRT_TV]:
            output_name += f"_r{rounding:.2f}"

        if scanline_strength > 0:
            output_name += f"_ssz{scanline_size:.2f}"

        if scanline_strength > 0:
            output_name += f"_sb{scanline_blur:.2f}"

        if scanline_strength > 0:
            output_name += f"_sst{scanline_strength:.2f}"

        if bloom_size > 0 and bloom_strength > 0:
            output_name += f"_bls{bloom_strength:.2f}"

        if grid_strength > 0:
            output_name += f"_gs{grid_strength:.2f}"

        if pixelate:
            output_name += "_pxt"

        output_name += ".png"

        # Save
        image.save(os.path.join(save_location, output_name))
        print(f"Saved: {output_name}")

    def test_make_images(self):
        # Clear the directory
        if os.path.exists(save_dir) and os.path.isdir(save_dir):
            shutil.rmtree(save_dir)

        os.mkdir(save_dir)

        # Compute the images
        for px_size in [130, 66, 33, 22, 10, 7, 6, 5, 4, 3]:
            for pixel_aspect in [0.33, 1, 3]:
                for direction in [filters.Direction.VERTICAL, filters.Direction.HORIZONTAL]:
                    for scanline_spacing in [0.33, 1.0, 3]:
                        self.make_image(
                            save_location=save_dir,
                            screen_type=filters.ScreenType.CRT_TV,
                            pixel_size=px_size,
                            pixel_padding=0.25,
                            direction=direction,
                            washout=0.5,
                            blur=0.5,
                            bloom_size=0.5,
                            pixel_aspect=pixel_aspect,
                            rounding=0.5,
                            scanline_spacing=scanline_spacing,
                            scanline_size=0.75,
                            scanline_blur=0.25,
                            scanline_strength=1,
                            bloom_strength=1,
                            grid_strength=1,
                            pixelate=True,
                            output_scale=4
                        )

                        self.make_image(
                            save_location=save_dir,
                            screen_type=filters.ScreenType.CRT_MONITOR,
                            pixel_size=px_size,
                            pixel_padding=0.1,
                            direction=direction,
                            washout=0.5,
                            blur=0.75,
                            bloom_size=0.5,
                            pixel_aspect=pixel_aspect,
                            scanline_spacing=scanline_spacing,
                            scanline_size=0.75,
                            scanline_blur=0.25,
                            scanline_strength=1,
                            bloom_strength=1,
                            grid_strength=1,
                            pixelate=True,
                            output_scale=4
                        )

                    self.make_image(
                        save_location=save_dir,
                        screen_type=filters.ScreenType.LCD,
                        pixel_size=px_size,
                        pixel_padding=0.25,
                        direction=direction,
                        washout=0.1,
                        blur=0,
                        bloom_size=0.5,
                        pixel_aspect=pixel_aspect,
                        rounding=0,
                        scanline_strength=0,
                        bloom_strength=1,
                        grid_strength=1,
                        pixelate=True,
                        output_scale=4
                    )


if __name__ == '__main__':
    unittest.main()



import os
import math
import glob
from PIL import Image


# General purpose text input stripper
def strip_all(input_text):
    return input_text.strip().strip("\n").strip("\r").strip()


# Makes sure a string represents a valid, existing file
# This can be used with argparse as a valid argument type
def file_path(string):
    if os.path.isfile(string):
        return string
    else:
        raise FileNotFoundError(string)


# A short function to clip a value to a range
def clip(value, min_value, max_value):
    return min(max(value, min_value), max_value)


# A function to give absolute bounding coordinates for a specified center point and object size
def get_centered_dimensions(center, size):
    start = (
        center[0] - (size[0] / 2),
        center[1] - (size[1] / 2),
    )
    end = (
        center[0] + (size[0] / 2),
        center[1] + (size[1] / 2),
    )

    return start, end


# A function to round to a given "division", like 0.5 instead of the default 1.0
def round_to_division(value, division=1.0):
    inverse = 1/division
    return round(value * inverse) / inverse


# Tile a PIL Image to fit a given frame size
def tile_image(image_tile, size, background_color=(0, 0, 0), count=None):
    new_image = Image.new(image_tile.mode, size, color=background_color)

    if count is None:
        # Position pastes based on tile size
        x_count = math.ceil(new_image.width / image_tile.width)
        y_count = math.ceil(new_image.height / image_tile.height)

        for x in range(x_count):
            for y in range(y_count):
                new_image.paste(image_tile, (x * image_tile.width, y * image_tile.height))
    else:
        # Position pastes based on count tuple (can be a float)
        x_count, y_count = count
        x_count_int = math.ceil(x_count)
        y_count_int = math.ceil(y_count)
        for x in range(x_count_int):
            for y in range(y_count_int):
                tile_start = (
                        round((x / x_count) * new_image.width),
                        round((y / y_count) * new_image.height)
                    )
                tile_end = (
                    round(((x + 1) / x_count) * new_image.width),
                    round(((y + 1) / y_count) * new_image.height)
                )

                target_size = (
                    tile_end[0] - tile_start[0],
                    tile_end[1] - tile_start[1]
                )

                if image_tile.size == target_size:
                    this_tile = image_tile
                else:
                    this_tile = image_tile.resize(target_size, resample=Image.Resampling.LANCZOS)

                new_image.paste(this_tile, tile_start)

    return new_image


# Mix a PIL image with white
def mix_color_with_image(image, color, factor):
    if factor <= 0:
        return image

    white_image = Image.new(image.mode, image.size, color=color)

    if factor >= 1:
        return white_image

    new_image = Image.blend(image, white_image, factor)

    return new_image


# If a value is out of range, raise an exception
def assert_value_in_range(value, minimum=None, maximum=None, message=None):
    raise_error = False
    if minimum is not None:
        if value < minimum:
            raise_error = True
    if maximum is not None:
        if value > maximum:
            raise_error = True

    if raise_error:
        if message is None:
            if minimum is not None and maximum is not None:
                raise ValueError(f"Expected a value between {minimum} and {maximum}, got {value}")
            elif minimum is not None:
                raise ValueError(f"Expected a value no less than {minimum}, got {value}")
            elif maximum is not None:
                raise ValueError(f"Expected a value no more than {maximum}, got {value}")
        else:
            raise ValueError(message.format(min=minimum, max=maximum, val=value))


# Helper function to get info from a sequenced image filename
def parse_sequenced_image_name(image_file):
    main_path, image_filename = os.path.split(image_file)
    file_basename, ext = os.path.splitext(image_filename)
    number = ""

    for x in file_basename[::-1]:
        if x.isdigit():
            number += x
        else:
            break

    if len(number) > 0:
        return {"error": None,
                "digits": len(number),
                "image_number": int(number),
                "prefix": os.path.join(main_path, "".join(file_basename.split(number)[:-1])),
                "ext": ext}
    else:
        return {"error": "No trailing digits found."}


def get_all_images_in_sequence(image_file):
    info = parse_sequenced_image_name(image_file)

    if info["error"] is None:
        return {
            "files": glob.glob(info["prefix"] + ("[0-9]" * info["digits"]) + info["ext"]),
            "digits": info["digits"],
            "ext": info["ext"],
            "prefix": info["prefix"]
        }
    else:
        return None
